from collections.abc import Awaitable, Callable
from inspect import iscoroutinefunction
from typing import ParamSpec, TypeGuard, TypeVar

_P = ParamSpec("_P")
_R = TypeVar("_R")


def is_async_func(func: Callable[_P, _R | Awaitable[_R]]) -> TypeGuard[Callable[_P, Awaitable[_R]]]:
    """
    Check if a function is an asynchronous function.

    Args
    ----
        func (`Callable`): The function to check.

    Returns
    -------
        `True` if the function is an asynchronous, `False` otherwise.
    """
    return callable(func) and iscoroutinefunction(func)


def is_sync_func(func: Callable[_P, _R | Awaitable[_R]]) -> TypeGuard[Callable[_P, _R]]:
    """
    Check if a function is synchronous.

    Args
    ----
        func (`Callable`): The function to check.

    Returns
    -------
        `True` if the function is synchronous, `False` otherwise.
    """
    return callable(func) and not iscoroutinefunction(func)



from typing import TYPE_CHECKING, Any, Literal, NamedTuple, overload

from ..internal import _InternalView  # noqa: TID252
from ..state import State  # noqa: TID252
from ..utils import wait_tasks_by_name  # noqa: TID252

if TYPE_CHECKING:
    from collections.abc import Generator

    from discord import Message

    from ..view import View, ViewObject  # noqa: TID252
    from .type import ViewObjectDictWithAttachment, ViewObjectDictWithFiles


class ViewResult(NamedTuple):
    """
    ViewResult is a named tuple representing the result of the view.

    Parameters
    ----------
    NamedTuple : `ViewResult`
        The result of the view.
    """

    timed_out: bool
    states: dict[str, Any]


class ViewController:
    """ViewController is a class that controls the view."""

    def __init__(self, view: "View", *, timeout: float | None = 180) -> None:
        self.__view = view
        view._controller = self  # noqa: SLF001
        self.__raw_view = _InternalView(timeout=timeout, on_error=self.__view.on_error, on_timeout=self.__view.on_timeout)
        self.__message: "Message | None" = None

    @property
    def message(self) -> "Message | None":
        """
        return attached message with the View.

        Returns
        -------
        `discord.Message | None`
            The attached message. None if the View is not sent yet.
        """
        return self.__message

    @message.setter
    def message(self, value: "Message | None") -> None:
        self.__message = value

    async def send(self) -> None:
        """
        Send the view to the channel.

        Raises
        ------
        NotImplementedError
            If this method is not implemented in subclasses.
        """
        # implement this in subclasses
        raise NotImplementedError

    async def sync(self) -> None:
        """Sync the message with current view."""
        if self.message is None:
            return

        # maybe validation for self.__view is needed
        d = self._process_view_for_discord("attachment")
        self.message = await self.message.edit(**d)

    def stop(self) -> None:
        """Stop the view and return the state of all states in the view."""
        self.__raw_view.stop()

    async def wait(self) -> ViewResult:
        """
        Wait for the view to stop and return the state of all states in the view.

        Returns
        -------
        `ViewResult(NamedTuple)`
            The result of the view.

            `timed_out` is True if the view timed out and False otherwise. same as `discord.ui.View.wait`.

            `states` is a dictionary of all states in the view.
        """
        is_timed_out = await self.__raw_view.wait()

        if is_timed_out:
            # this is got from discord.ui.View._dispatch_timeout()
            timeout_task_name = f"discord-ui-view-timeout-{self.__raw_view.id}"
            await wait_tasks_by_name([timeout_task_name])

        d = {}
        for key, state in self._get_all_state_in_view():
            d[key] = state.get_state()
        return ViewResult(is_timed_out, d)

    def _get_all_state_in_view(self) -> "Generator[tuple[str, State[Any]], None, None]":
        for k, v in self.__view.__dict__.items():
            if isinstance(v, State):
                yield k, v

    @overload
    def _process_view_for_discord(self, mode: Literal["attachment"]) -> "ViewObjectDictWithAttachment":
        ...

    @overload
    def _process_view_for_discord(self, mode: Literal["files"]) -> "ViewObjectDictWithFiles":
        ...

    def _process_view_for_discord(
        self,
        mode: Literal["attachment", "files"],
    ) -> "ViewObjectDictWithAttachment | ViewObjectDictWithFiles":
        """
        _process_view_for_discord is a helper function to process the view for Discord.

        Parameters
        ----------
        mode : Literal[&quot;attachment&quot;, &quot;files&quot;]
            The mode to process the view for Discord.

            If the mode is `attachment`, ViewObject.files will be put into the `attachments` key.

            If the mode is `files`, ViewObject.files will be put into the `files` key.

        Returns
        -------
        ViewObjectDictWithAttachment | ViewObjectDictWithFiles
            The processed view dictionary.
            This can be passed to `discord.abc.Messageable.send` or `discord.abc.Messageable.edit` and etc
            as unpacked keyword arguments.
        """
        view_object: "ViewObject" = self.__view.render()

        # implicitly clear view every time see:#54
        v = self.__raw_view
        v.clear_items()
        if view_object.components:
            for child in view_object.components:
                v.add_item(child)

        if mode == "attachment":
            # implicitly clear items every time see:#54

            return {
                "content": view_object.content,
                "embeds": view_object.embeds or [],
                "view": v,
                "attachments": view_object.files or [],
            }

        return {
            "content": view_object.content,
            "embeds": view_object.embeds or [],
            "view": v,
            "files": view_object.files or [],
        }



from .callback import (
    ChannelSelectCallback,
    ChannelSelectSyncCallback,
    #
    InteractionCallback,
    InteractionSyncCallback,
    #
    MentionableSelectCallback,
    MentionableSelectSyncCallback,
    #
    ModalCallback,
    ModalSyncCallback,
    #
    RoleSelectCallback,
    RoleSelectSyncCallback,
    #
    SelectCallback,
    SelectSyncCallback,
    #
    UserSelectCallback,
    UserSelectSyncCallback,
)
from .view import ViewErrorHandler, ViewTimeoutHandler

__all__ = [
    "InteractionCallback",
    "InteractionSyncCallback",
    "SelectCallback",
    "SelectSyncCallback",
    "ChannelSelectCallback",
    "ChannelSelectSyncCallback",
    "RoleSelectCallback",
    "RoleSelectSyncCallback",
    "MentionableSelectCallback",
    "MentionableSelectSyncCallback",
    "UserSelectCallback",
    "UserSelectSyncCallback",
    "ModalCallback",
    "ModalSyncCallback",
    "ViewErrorHandler",
    "ViewTimeoutHandler",
]



from .controller import ViewController
from .interaction_controller import InteractionController
from .messageable_controller import MessageableController

__all__ = ["InteractionController", "MessageableController", "ViewController"]



from typing import TYPE_CHECKING, Literal

from discord import TextStyle, ui
from typing_extensions import NotRequired, Required, TypedDict

from ..utils import call_any_function, is_sync_func  # noqa: TID252

if TYPE_CHECKING:
    from discord import Interaction

    from ..types import ModalCallback, ModalSyncCallback  # noqa: TID252


class TextInputStyle(TypedDict):
    """TextInputStyle is a TypedDict that represents the style of a text input."""

    field: Required[Literal["short", "long"]]
    placeholder: NotRequired[str | None]
    default: NotRequired[str | None]
    row: NotRequired[Literal[0, 1, 2, 3, 4]]


class TextInputConfig(TypedDict, total=False):
    """TextInputConfig is a TypedDict that represents the config of a text input."""

    required: bool
    min_length: int | None
    max_length: int | None


class TextInput(ui.TextInput):
    """
    TextInput is a class that represents a text input.

    This class has compatibility with the `discord.ui.TextInput` class.

    """

    def __init__(
        self,
        label: str,
        /,
        *,
        style: TextInputStyle,
        config: TextInputConfig,
        custom_id: str | None = None,
    ) -> None:
        __d = {
            "label": label,
            "style": TextStyle[style.get("field", "short")],
            "placeholder": style.get("placeholder", None),
            "default": style.get("default", None),
            "required": config.get("required", False),
            "row": style.get("row", None),
            "min_length": config.get("min_length", None),
            "max_length": config.get("max_length", None),
        }
        if custom_id:
            __d["custom_id"] = custom_id
        super().__init__(**__d)


class Modal(ui.Modal):
    """
    Modal is a class that represents a modal.

    This class has compatibility with the `discord.ui.Modal` class.
    """

    def __init__(  # noqa: PLR0913
        self,
        *,
        title: str,
        inputs: list[TextInput],
        timeout: float | None = None,
        custom_id: str | None = None,
        on_submit: "ModalCallback | ModalSyncCallback | None" = None,
    ) -> None:
        __d = {
            "title": title,
            "timeout": timeout,
        }
        if custom_id:
            __d["custom_id"] = custom_id
        self.__callback_fn = on_submit
        self.__inputs = inputs
        super().__init__(**__d)
        for _in in self.__inputs:
            self.add_item(_in)

    async def on_submit(self, interaction: "Interaction") -> None:  # noqa: D102
        if self.__callback_fn is None:
            await interaction.response.defer()
            return

        if is_sync_func(self.__callback_fn):
            await interaction.response.defer()

        await call_any_function(self.__callback_fn, interaction, {i.label: i.value for i in self.__inputs})



from asyncio import iscoroutinefunction
from collections.abc import Callable
from typing import ParamSpec, TypeVar

P = ParamSpec("P")
R = TypeVar("R")

__all__ = [
    "call_any_function",
]


async def call_any_function(fn: Callable[P, R], *args: P.args, **kwargs: P.kwargs) -> R:
    """
    Call a function, whether it is a coroutine or not.

    This function is used internally to call functions that may or may not be coroutines.

    Parameters
    ----------
    fn : Callable[P, R]
        The function to call.
    *args : P.args
        The positional arguments to pass to the function.
    **kwargs : P.kwargs
        The keyword arguments to pass to the function.

    Returns
    -------
    R
        The return value of the function.
    """
    if iscoroutinefunction(fn):
        return await fn(*args, **kwargs)
    return fn(*args, **kwargs)



import asyncio
from typing import TYPE_CHECKING

from discord import Embed, File, ui
from pydantic import BaseModel, Field

from .utils import get_logger

if TYPE_CHECKING:
    from discord import Interaction

    from .controller import ViewController


__all__ = [
    "View",
    "ViewObject",
]


class ViewObject(BaseModel):
    """
    A class representing a view object that can be sent as a message in Discord.

    Attributes
    ----------
    content : `str`
        The content of the message.
    embeds : `list[discord.Embed] | None`
        A list of embeds to be included in the message.
    files : `list[discord.File] | None`
        A list of files to be included in the message.
    components : `list[discord.ui.Item] | None`
        A list of UI components to be included in the message.
    """

    content: str = Field(default="")
    embeds: list[Embed] | None = Field(default=None)
    files: list[File] | None = Field(default=None)
    components: list[ui.Item] | None = Field(default=None)

    model_config = {"arbitrary_types_allowed": True}


class View:
    """
    The View class represents the user interface of the UI. It is responsible for rendering the UI and.

    handling user interactions.

    Attributes
    ----------
    message : `discord.Message | None`
        The message containing the UI. None if the UI has not been sent yet.

    Methods
    -------
    render() -> `ViewObject`:
        Renders the UI and returns a `ViewObject` representing the UI.
    sync() -> `None`:
        Synchronizes the view with the controller.
    stop() -> `None`:
        Stops the view.
    on_error(interaction: `discord.Interaction`, error: `Exception`, item: `discord.ui.Item`) -> `None`:
        Called when an error occurs in the view.
    on_timeout() -> `None`:
        Called when the view times out.
    """

    def __init__(
        self,
        loop: asyncio.AbstractEventLoop | None = None,
    ) -> None:
        self._loop = loop or asyncio.get_event_loop()
        self._controller: "ViewController | None" = None
        self.__logger = get_logger(__name__)

    def render(self) -> ViewObject:
        """
        Render the view and returns a ViewObject. This method is called by `Controller`.

        Returns
        -------
        ViewObject: The rendered view.
        """
        return ViewObject()

    def sync(self) -> None:
        """Synchronize the view with the controller. This method is called by `State` when its value changes."""
        if self._controller:
            self._loop.create_task(self._controller.sync())
        else:
            self.__logger.warning("Controller is not set")

    def stop(self) -> None:
        """Stop the view. This method is called by child components implicitly."""
        if self._controller:
            self._controller.stop()
        else:
            self.__logger.warning("Controller is not set")

    async def on_error(self, interaction: "Interaction", error: Exception, item: "ui.Item") -> None:
        """
        on_error is called when an error occurs in the view.

        Parameters
        ----------
        interaction : `discord.Interaction`
            The interaction that caused the error.
        error : `Exception`
            The error that occurred.
        item : `discord.ui.Item`
            The item that caused the error.
        """

    async def on_timeout(self) -> None:
        """on_timeout is called when the view times out."""



from collections.abc import Generator
from typing import TypeVar

_T = TypeVar("_T")


def chunks(iterable: list[_T], size: int) -> Generator[list[_T], None, None]:
    """Yield successive chunks from iterable of size."""
    for i in range(0, len(iterable), size):
        yield iterable[i : i + size]



from collections.abc import Awaitable, Callable
from typing import TypeAlias

import discord

__all__ = [
    "ViewErrorHandler",
    "ViewTimeoutHandler",
]

ViewErrorHandler: TypeAlias = Callable[[discord.Interaction, Exception, discord.ui.Item], Awaitable[None]]
ViewTimeoutHandler: TypeAlias = Callable[[], Awaitable[None]]



from collections.abc import Awaitable, Callable
from typing import TypeAlias

import discord
from discord.app_commands import AppCommandChannel, AppCommandThread

__all__ = [
    "InteractionCallback",
    "SelectCallback",
    "ChannelSelectCallback",
    "RoleSelectCallback",
    "MentionableSelectCallback",
    "UserSelectCallback",
    "ModalCallback",
]

# InteractionCallback
InteractionCallback: TypeAlias = Callable[[discord.Interaction], Awaitable[None]]
InteractionSyncCallback: TypeAlias = Callable[[discord.Interaction], None]

# SelectCallback
SelectCallback: TypeAlias = Callable[[discord.Interaction, list[str]], Awaitable[None]]
SelectSyncCallback: TypeAlias = Callable[[discord.Interaction, list[str]], None]

ChannelSelectCallback: TypeAlias = Callable[
    [discord.Interaction, list[AppCommandChannel | AppCommandThread]],
    Awaitable[None],
]
ChannelSelectSyncCallback: TypeAlias = Callable[
    [discord.Interaction, list[AppCommandChannel | AppCommandThread]],
    None,
]

RoleSelectCallback: TypeAlias = Callable[
    [discord.Interaction, list[discord.Role]],
    Awaitable[None],
]
RoleSelectSyncCallback: TypeAlias = Callable[
    [discord.Interaction, list[discord.Role]],
    None,
]

MentionableSelectCallback: TypeAlias = Callable[
    [discord.Interaction, list[discord.Role | discord.Member | discord.User]],
    Awaitable[None],
]
MentionableSelectSyncCallback: TypeAlias = Callable[
    [discord.Interaction, list[discord.Role | discord.Member | discord.User]],
    None,
]

UserSelectCallback: TypeAlias = Callable[[discord.Interaction, list[discord.User | discord.Member]], Awaitable[None]]
UserSelectSyncCallback: TypeAlias = Callable[[discord.Interaction, list[discord.User | discord.Member]], None]


# ModalCallback
ModalCallback: TypeAlias = Callable[[discord.Interaction, dict[str, str]], Awaitable[None]]
ModalSyncCallback: TypeAlias = Callable[[discord.Interaction, dict[str, str]], None]



import asyncio
from collections.abc import Callable
from typing import TYPE_CHECKING, Any, Generic, TypeVar

from .utils import get_logger

if TYPE_CHECKING:
    from .view import View

T = TypeVar("T", bound=Any)

__all__ = [
    "State",
    # "use_state",
]


class State(Generic[T]):
    """
    A class representing a state with a generic type T.

    Methods
    -------
    __call__() -> `T`:
        Return the current value of the state.
    get_state() -> `T`:
        Return the current value of the state.
    set_state(new_value: `T | Callable[[T], T]`) -> `None`:
        Set the current value of the state to the new value.
    """

    def __init__(self, initial_value: T, view: "View", /, *, loop: asyncio.AbstractEventLoop | None = None) -> None:
        self.__initial_value: T = initial_value
        self.__current_value: T = initial_value
        self.__previous_value: T | None = None

        self._loop = loop or asyncio.get_event_loop()
        self._view = view
        self._logger = get_logger(self.__class__.__name__)

    def __call__(self) -> T:
        """
        Return the current value of the state. equivalent to `State.get_state()`.

        Returns
        -------
        `T`
            The current value of the state.
        """
        return self.get_state()

    @property
    def _current_value(self) -> T:
        """
        property: The current value of the state.

        Returns
        -------
        T
            The current value of the state.
        """
        return self.__current_value

    @_current_value.setter
    def _current_value(self, new_value: T) -> None:
        """
        Set the current value of the state to the new value.

        This method automatically sets the previous value of the state to the current value of the state.

        Parameters
        ----------
        new_value : T
            The new value of the state.
        """
        self.__previous_value = self.__current_value
        self.__current_value = new_value

    def get_state(self) -> T:
        """
        Return the current value of the state.

        Returns
        -------
        `T`
            The current value of the state.
        """
        return self._current_value

    def set_state(self, new_value: T | Callable[[T], T]) -> None:
        """
        Set the current value of the state to the new value.

        After the state is changed, this method calls `View.sync()` to synchronize the view with the controller.

        Parameters
        ----------
        new_value : `T | Callable[[T], T]`
            The new value of the state. If the type is `Callable[[T], T]`, the callable is called with the current
            value of the state and the return value is used as the new value of the state.
        """
        _new_value: T = self._current_value

        if isinstance(new_value, Callable):
            try:
                r = new_value(self._current_value)
            except Exception:
                self._logger.exception("Error while executing callable")
            else:
                if not isinstance(r, type(self._current_value)):
                    self._logger.warning("Callable returned value of different type")
                else:
                    _new_value = r
        else:
            _new_value = new_value

        msg = f"State changed: {self._current_value} -> {_new_value}"
        self._logger.debug(msg)
        self._current_value = _new_value
        self.__sync()

    def revert_state(self) -> None:
        """
        Revert the current value of the state to the previous value.

        After the state is changed, this method calls `View.sync()` to synchronize the view with the controller.
        """
        if self.__previous_value is not None:
            self._current_value = self.__previous_value
            self.__sync()
        else:
            self._logger.warning("No previous value")

    def __sync(self) -> None:
        if self._view:
            self._view.sync()
        else:
            self._logger.warning("View is not set")


# class UseStateTuple(NamedTuple, Generic[T]):
#     state: State[T]
#     set_state: Callable[[T | Callable[[T], T]], None]


# def use_state(
#     initial_value: T,
#     view: "View",
#     /,
#     *,
#     loop: asyncio.AbstractEventLoop | None = None,
# ) -> UseStateTuple[T]:
#     s = State[T](initial_value, view, loop=loop)
#     return UseStateTuple(state=s, set_state=s.set_state)



from .button import Button, LinkButton
from .modal import Modal, TextInput
from .select import ChannelSelect, MentionableSelect, RoleSelect, Select, SelectOption, UserSelect

__all__ = [
    "Button",
    "LinkButton",
    "Modal",
    "TextInput",
    "ChannelSelect",
    "MentionableSelect",
    "RoleSelect",
    "Select",
    "SelectOption",
    "UserSelect",
]



from typing import TYPE_CHECKING

from discord import ui

if TYPE_CHECKING:
    from discord import Interaction

    from ..types import ViewErrorHandler, ViewTimeoutHandler  # noqa: TID252

__all__ = [
    "_InternalView",
]


class _InternalView(ui.View):
    def __init__(
        self,
        *,
        timeout: float | None = 180,
        on_error: "ViewErrorHandler | None" = None,
        on_timeout: "ViewTimeoutHandler | None" = None,
    ) -> None:
        super().__init__(timeout=timeout)
        self.__on_error = on_error
        self.__on_timeout = on_timeout

    async def on_error(self, interaction: "Interaction", error: Exception, item: ui.Item) -> None:
        if self.__on_error:
            await self.__on_error(interaction, error, item)

        await super().on_error(interaction, error, item)

    async def on_timeout(self) -> None:
        if self.__on_timeout:
            await self.__on_timeout()

        await super().on_timeout()



from typing import TYPE_CHECKING

from discord import CategoryChannel, ForumChannel

from .controller import ViewController

if TYPE_CHECKING:
    from discord import Interaction

    from ..view import View  # noqa: TID252


class InteractionController(ViewController):
    """InteractionController is a class that controls the view with `discord.abc.Messageable`."""

    def __init__(
        self,
        view: "View",
        *,
        interaction: "Interaction",
        timeout: float | None = 180,
        ephemeral: bool = False,
    ) -> None:
        super().__init__(view, timeout=timeout)
        self.__interaction = interaction
        self.__ephemeral = ephemeral

    async def send(self) -> None:
        """Send the view to the channel."""
        target = self.__interaction
        view_kwargs = self._process_view_for_discord("files")

        if target.is_expired():
            if target.channel is not None and not isinstance(target.channel, CategoryChannel | ForumChannel):
                self.message = await target.channel.send(**view_kwargs)
            return

        if target.response.is_done():
            self.message = await target.followup.send(**view_kwargs, ephemeral=self.__ephemeral, wait=True)
            return

        await target.response.send_message(**view_kwargs, ephemeral=self.__ephemeral)
        self.message = await target.original_response()
        return



from typing import TYPE_CHECKING, Literal

from discord import Emoji, PartialEmoji, ui
from discord import SelectOption as _SelectOption
from typing_extensions import NotRequired, Required, TypedDict

from ..utils import call_any_function, is_sync_func  # noqa: TID252

if TYPE_CHECKING:
    from discord import ChannelType, Interaction

    from ..types import (  # noqa: TID252
        ChannelSelectCallback,
        ChannelSelectSyncCallback,
        MentionableSelectCallback,
        MentionableSelectSyncCallback,
        RoleSelectCallback,
        RoleSelectSyncCallback,
        SelectCallback,
        SelectSyncCallback,
        UserSelectCallback,
        UserSelectSyncCallback,
    )


class SelectStyle(TypedDict, total=False):
    """SelectStyle is a TypedDict that represents the style of a select."""

    disabled: bool
    placeholder: str | None
    row: Literal[0, 1, 2, 3, 4]


# class SelectOption(BaseModel):
#     """
#     SelectOption is a class that represents a select option.

#     This class extends the `pydantic.BaseModel` class and has validation.
#     """

#     label: str = Field(min_length=1, max_length=100)
#     value: str | None = Field(default=None, min_length=1, max_length=100)
#     description: str | None = Field(default=None, min_length=1, max_length=100)
#     emoji: str | Emoji | PartialEmoji | None = Field(default=None)
#     selected_by_default: bool = Field(default=False)

#     model_config = {"arbitrary_types_allowed": True}


class SelectOption(TypedDict):
    """SelectOption is a TypedDict that represents a select option."""

    label: Required[str]
    value: NotRequired[str | None]
    description: NotRequired[str | None]
    emoji: NotRequired[str | Emoji | PartialEmoji | None]
    default: NotRequired[bool]


class SelectConfigBase(TypedDict, total=False):
    """SelectConfigBase is a TypedDict that represents the config of a select."""

    min_values: int | None
    max_values: int | None


class SelectConfig(SelectConfigBase):
    """SelectConfig is a class that represents the config of a select."""


class ChannelSelectConfig(SelectConfigBase):
    """ChannelSelectConfig is a class that represents the config of a channel select."""

    channel_types: "list[ChannelType]"


class RoleSelectConfig(SelectConfigBase):
    """RoleSelectConfig is a class that represents the config of a role select."""


class MentionableSelectConfig(SelectConfigBase):
    """MentionableSelectConfig is a class that represents the config of a mentionable select."""


class UserSelectConfig(SelectConfigBase):
    """UserSelectConfig is a class that represents the config of a user select."""


class Select(ui.Select):
    """
    Select is a class that represents a select.

    This class has compatibility with the `discord.ui.Select` class.
    """

    def __init__(  # noqa: PLR0913
        self,
        *,
        config: SelectConfig,
        style: SelectStyle,
        options: list[SelectOption],
        custom_id: str | None = None,
        on_select: "SelectCallback | SelectSyncCallback | None" = None,
    ) -> None:
        __disabled = style.get("disabled", False)
        __placeholder = style.get("placeholder", None)
        __row = style.get("row", None)
        __d = {
            "disabled": __disabled,
            "placeholder": __placeholder,
            "row": __row,
            "min_values": config.get("min_values", None),
            "max_values": config.get("max_values", None),
            "options": [
                _SelectOption(
                    label=(_label := option.get("label", "")),
                    value=option.get("value", None) or _label,
                    description=option.get("description", None),
                    emoji=option.get("emoji", None),
                    default=option.get("default", False),
                )
                for option in options
            ],
        }
        if custom_id:
            __d["custom_id"] = custom_id
        self.__callback_fn = on_select
        super().__init__(**__d)

    async def callback(self, interaction: "Interaction") -> None:  # noqa: D102
        if self.__callback_fn is None:
            await interaction.response.defer()
            return

        if is_sync_func(self.__callback_fn):
            await interaction.response.defer()

        await call_any_function(self.__callback_fn, interaction, self.values)


class ChannelSelect(ui.ChannelSelect):
    """
    ChannelSelect is a class that represents a channel select.

    This class has compatibility with the `discord.ui.ChannelSelect` class.
    """

    def __init__(
        self,
        *,
        config: ChannelSelectConfig,
        style: SelectStyle,
        custom_id: str | None = None,
        on_select: "ChannelSelectCallback| ChannelSelectSyncCallback | None" = None,
    ) -> None:
        __disabled = style.get("disabled", False)
        __placeholder = style.get("placeholder", None)
        __row = style.get("row", None)
        __d = {
            "disabled": __disabled,
            "placeholder": __placeholder,
            "row": __row,
            "min_values": config.get("min_values", None),
            "max_values": config.get("max_values", None),
            "channel_types": config["channel_types"],
        }
        if custom_id:
            __d["custom_id"] = custom_id
        self.__callback_fn = on_select
        super().__init__(**__d)

    async def callback(self, interaction: "Interaction") -> None:  # noqa: D102
        if self.__callback_fn is None:
            await interaction.response.defer()
            return

        if is_sync_func(self.__callback_fn):
            await interaction.response.defer()

        await call_any_function(self.__callback_fn, interaction, self.values)


class RoleSelect(ui.RoleSelect):
    """
    RoleSelect is a class that represents a role select.

    This class has compatibility with the `discord.ui.RoleSelect` class.
    """

    def __init__(
        self,
        *,
        config: RoleSelectConfig,
        style: SelectStyle,
        custom_id: str | None = None,
        on_select: "RoleSelectCallback | RoleSelectSyncCallback | None" = None,
    ) -> None:
        __disabled = style.get("disabled", False)
        __placeholder = style.get("placeholder", None)
        __row = style.get("row", None)
        __d = {
            "disabled": __disabled,
            "placeholder": __placeholder,
            "row": __row,
            "min_values": config.get("min_values", None),
            "max_values": config.get("max_values", None),
        }
        if custom_id:
            __d["custom_id"] = custom_id
        self.__callback_fn = on_select
        super().__init__(**__d)

    async def callback(self, interaction: "Interaction") -> None:  # noqa: D102
        if self.__callback_fn is None:
            await interaction.response.defer()
            return

        if is_sync_func(self.__callback_fn):
            await interaction.response.defer()

        await call_any_function(self.__callback_fn, interaction, self.values)


class MentionableSelect(ui.MentionableSelect):
    """
    MentionableSelect is a class that represents a mentionable select.

    This class has compatibility with the `discord.ui.MentionableSelect` class.

    """

    def __init__(
        self,
        *,
        config: MentionableSelectConfig,
        style: SelectStyle,
        custom_id: str | None = None,
        on_select: "MentionableSelectCallback | MentionableSelectSyncCallback | None" = None,
    ) -> None:
        __disabled = style.get("disabled", False)
        __placeholder = style.get("placeholder", None)
        __row = style.get("row", None)
        __d = {
            "disabled": __disabled,
            "placeholder": __placeholder,
            "row": __row,
            "min_values": config.get("min_values", None),
            "max_values": config.get("max_values", None),
        }
        if custom_id:
            __d["custom_id"] = custom_id
        self.__callback_fn = on_select
        super().__init__(**__d)

    async def callback(self, interaction: "Interaction") -> None:  # noqa: D102
        if self.__callback_fn is None:
            await interaction.response.defer()
            return

        if is_sync_func(self.__callback_fn):
            await interaction.response.defer()

        await call_any_function(self.__callback_fn, interaction, self.values)


class UserSelect(ui.UserSelect):
    """
    UserSelect is a class that represents a user select.

    This class has compatibility with the `discord.ui.UserSelect` class.
    """

    def __init__(
        self,
        *,
        config: UserSelectConfig,
        style: SelectStyle,
        custom_id: str | None = None,
        on_select: "UserSelectCallback | UserSelectSyncCallback | None" = None,
    ) -> None:
        __disabled = style.get("disabled", False)
        __placeholder = style.get("placeholder", None)
        __row = style.get("row", None)
        __d = {
            "disabled": __disabled,
            "placeholder": __placeholder,
            "row": __row,
            "min_values": config.get("min_values", None),
            "max_values": config.get("max_values", None),
        }
        if custom_id:
            __d["custom_id"] = custom_id
        self.__callback_fn = on_select
        super().__init__(**__d)

    async def callback(self, interaction: "Interaction") -> None:  # noqa: D102
        if self.__callback_fn is None:
            await interaction.response.defer()
            return

        if is_sync_func(self.__callback_fn):
            await interaction.response.defer()

        await call_any_function(self.__callback_fn, interaction, self.values)



from . import controller, pagination, types, ui
from .state import State
from .view import View, ViewObject

__all__ = [
    "controller",
    "pagination",
    "types",
    "ui",
    "State",
    "View",
    "ViewObject",
]



import asyncio
from typing import Any


def get_all_tasks(*, loop: asyncio.AbstractEventLoop | None = None, only_undone: bool = False) -> set[asyncio.Task]:
    """
    Return a set of all tasks running on the given event loop.

    Args:
        loop `(asyncio.AbstractEventLoop | None)`: The event loop to get tasks from.
        If None, the current running loop is used.
        only_undone `(bool)`: If True, only tasks that are not done are returned.

    Returns
    -------
        `set[asyncio.Task]`: A set of all tasks running on the given event loop.
    """
    _loop: asyncio.AbstractEventLoop | None = loop or safe_get_running_loop()

    if _loop is None:
        return set()

    tasks = asyncio.all_tasks(loop=_loop)

    if only_undone:
        return {t for t in tasks if not t.done()}

    return tasks


async def wait_tasks_by_name(names: list[str]) -> list[asyncio.Future[Any]]:
    """
    Wait for all tasks with the given names to complete.

    Args:
        names `(list[str])`: A list of task names to wait for.

    Returns
    -------
        `list[asyncio.Future[Any]]`: A list of futures representing the results of the completed tasks.
    """
    tasks = [t for t in get_all_tasks() if t.get_name() in names]
    return await asyncio.gather(*tasks, return_exceptions=True)


def safe_get_running_loop() -> asyncio.AbstractEventLoop | None:
    """
    Get the running loop safely.

    Returns
    -------
    `asyncio.AbstractEventLoop | None`
        The running loop. None if faced RuntimeError.
    """
    try:
        return asyncio.get_running_loop()
    except RuntimeError:
        return None



from .async_helper import get_all_tasks, wait_tasks_by_name
from .call import call_any_function
from .chunk import chunks
from .logger import get_logger
from .type_helper import is_async_func, is_sync_func

__all__ = [
    "chunks",
    "get_all_tasks",
    "wait_tasks_by_name",
    "call_any_function",
    "get_logger",
    "is_async_func",
    "is_sync_func",
]



from typing import TYPE_CHECKING, Generic, TypeVar

from typing_extensions import NotRequired, Required, TypedDict

from ..utils import chunks  # noqa: TID252

if TYPE_CHECKING:
    from discord import Interaction

    from ..view import View  # noqa: TID252


_T = TypeVar("_T")


class PaginatorConfig(TypedDict):
    """
    A paginator configuration.

    Parameters
    ----------
    TypedDict : _type_
        _description_
    """

    page_size: Required[int]
    initial_page: NotRequired[int]


class Paginator(Generic[_T]):
    """
    A paginator.

    Parameters
    ----------
    view : `View`
        The view to attach.

    source: `list[_T]`
        The source data. This can be any iterable.

    config: `PaginatorConfig`
        The paginator configuration.
    """

    def __init__(self, view: "View", *, source: list[_T], config: PaginatorConfig) -> None:
        self.__view = view
        self.__CHUNKS = list(chunks(source, config["page_size"]))
        self.__MAX_INDEX: int = len(self.__CHUNKS) - 1
        self.__current_index: int = c if (self._is_valid_index(c := (config.get("initial_page", 0)))) else 0

    @property
    def current_page(self) -> int:
        """
        Return the current page number.

        Returns
        -------
        int
            The current page number.
        """
        return self.__current_index + 1

    @property
    def max_page(self) -> int:
        """
        Return the maximum page number.

        Returns
        -------
        int
            The maximum page number.
        """
        return self.__MAX_INDEX + 1

    @property
    def at_first(self) -> bool:
        """
        Return whether the current page is the first page.

        Returns
        -------
        bool
            Whether the current page is the first page.
        """
        return self.__current_index == 0

    @property
    def at_last(self) -> bool:
        """
        Return whether the current page is the last page.

        Returns
        -------
        bool
            Whether the current page is the last page.
        """
        return self.__current_index == self.__MAX_INDEX

    def _is_valid_index(self, index: int) -> bool:
        return 0 <= index <= self.__MAX_INDEX

    def go_next(self, _: "Interaction") -> None:
        """Go to the next page. This method will call `View.sync`."""
        next_index = self.__current_index + 1
        if not self._is_valid_index(next_index):
            return

        self.__current_index = next_index
        self.__view.sync()

    def go_previous(self, _: "Interaction") -> None:
        """Go to the previous page. This method will call `View.sync`."""
        previous_index = self.__current_index - 1
        if not self._is_valid_index(previous_index):
            return

        self.__current_index = previous_index
        self.__view.sync()

    def go_first(self, _: "Interaction") -> None:
        """Go to the first page. This method will call `View.sync`."""
        if not self._is_valid_index(self.__current_index) or self.at_first:
            return

        self.__current_index = 0
        self.__view.sync()

    def go_last(self, _: "Interaction") -> None:
        """Go to the last page. This method will call `View.sync`."""
        if not self._is_valid_index(self.__current_index) or self.at_last:
            return

        self.__current_index = self.__MAX_INDEX
        self.__view.sync()

    @property
    def data(self) -> list[_T]:
        """
        Return the current page data.

        Returns
        -------
        _T
            The current page data.
        """
        return self.__CHUNKS[self.__current_index]



from typing import TYPE_CHECKING, Literal

from discord import ButtonStyle as _ButtonStyle
from discord import ui
from typing_extensions import NotRequired, Required, TypedDict

from ..utils import call_any_function, is_sync_func  # noqa: TID252

if TYPE_CHECKING:
    from discord import Emoji, Interaction, PartialEmoji

    from ..types import InteractionCallback, InteractionSyncCallback  # noqa: TID252


class ButtonStyle(TypedDict):
    """ButtonStyle is a TypedDict that represents the style of a button."""

    color: Required[Literal["blurple", "grey", "green", "red"]]
    disabled: NotRequired[bool]
    emoji: NotRequired["str | Emoji | PartialEmoji | None"]
    row: NotRequired[Literal[0, 1, 2, 3, 4]]


class Button(ui.Button):
    """
    Button is a class that represents a button.

    This class has compatibility with the `discord.ui.Button` class.
    """

    def __init__(
        self,
        label: str | None = None,
        /,
        *,
        style: ButtonStyle,
        custom_id: str | None = None,
        on_click: "InteractionCallback | InteractionSyncCallback | None" = None,
    ) -> None:
        __style = _ButtonStyle[style.get("color", "grey")]
        __disabled = style.get("disabled", False)
        __emoji = style.get("emoji", None)
        __row = style.get("row", None)
        self.__callback_fn = on_click
        super().__init__(
            style=__style,
            disabled=__disabled,
            emoji=__emoji,
            row=__row,
            label=label,
            custom_id=custom_id,
        )

    async def callback(self, interaction: "Interaction") -> None:  # noqa: D102
        if self.__callback_fn is None:
            await interaction.response.defer()
            return

        if is_sync_func(self.__callback_fn):
            await interaction.response.defer()

        await call_any_function(self.__callback_fn, interaction)


class LinkButton(ui.Button):
    """
    LinkButton is a class that represents a link button.

    This class has compatibility with the `discord.ui.Button` class.
    """

    def __init__(self, label: str | None = None, /, *, url: str, custom_id: str | None = None) -> None:
        super().__init__(
            style=_ButtonStyle.link,
            url=url,
            label=label,
            custom_id=custom_id,
        )

    async def callback(self, interaction: "Interaction") -> None:  # noqa: D102
        pass



from typing import TYPE_CHECKING

from typing_extensions import TypedDict

if TYPE_CHECKING:
    from discord import Embed, File, ui


class _ViewObjectDict(TypedDict, total=False):
    """
    ViewObjectDict is a type hint for the dictionary that is used to send a view to Discord.

    This can be passed to `discord.abc.Messageable.send` as unpacked keyword arguments.

    Accessing the keys are not recommended since the existence of the keys are not guaranteed.

    Example
    -------
    ```py
    d: ViewObjectDict = {
        "content": "Hello, world!",
        "embeds": [discord.Embed(title="Hello, world!")],
    }
    await ctx.send(**d)
    """

    content: str
    embeds: "list[Embed]"
    view: "ui.View"


class ViewObjectDictWithAttachment(_ViewObjectDict, total=False):  # noqa: D101
    attachments: "list[File]"


class ViewObjectDictWithFiles(_ViewObjectDict, total=False):  # noqa: D101
    files: "list[File]"



import logging
import logging.handlers

__all__ = [
    "get_logger",
]


def get_logger(name: str, level: str = "DEBUG") -> logging.Logger:  # name: __name__
    """
    Get a logger with a stream handler.

    Parameters
    ----------
    name : str
        logger name. Usually __name__.
    level : str, optional
        logging level, by default "DEBUG"

    Returns
    -------
    logging.Logger
        _description_
    """
    logger = logging.getLogger(name)
    stream_handler = logging.StreamHandler()

    # set format
    literal_formatter = logging.Formatter("%(asctime)s:%(levelname)s:\n%(name)s:%(message)s")
    stream_handler.setFormatter(literal_formatter)

    # set level
    logger.setLevel(level)
    stream_handler.setLevel(level)

    # add handler
    if not logger.hasHandlers():
        logger.addHandler(stream_handler)

    return logger



from typing import TYPE_CHECKING

from .controller import ViewController

if TYPE_CHECKING:
    import discord

    from ..view import View  # noqa: TID252


class MessageableController(ViewController):
    """MessageableController is a class that controls the view with `discord.abc.Messageable`."""

    def __init__(self, view: "View", *, messageable: "discord.abc.Messageable", timeout: float | None = 180) -> None:
        super().__init__(view, timeout=timeout)
        self.__messageable = messageable

    async def send(self) -> None:
        """Send the view to the channel."""
        target = self.__messageable
        view_kwargs = self._process_view_for_discord("files")

        self.message = await target.send(**view_kwargs)
